import datetime
from urllib.parse import quote
from zoneinfo import ZoneInfo

import pytest

import feedgenerator

AWARE_DATE = datetime.datetime(2016,8,11,0,0,0,0,tzinfo=ZoneInfo("America/New_York"))
AWARE_DATE_UTC = datetime.datetime(2016,8,11,0,0,0,0,tzinfo=ZoneInfo("UTC"))
NAIVE_DATE = datetime.datetime(2016,8,11,0,0,0,0)

FIXT_FEED = dict(
    title="Poynter E-Media Tidbits",
    link="http://www.poynter.org/column.asp?id=31",
    description="""A group Weblog by the sharpest minds in online media/journalism/publishing.
    Umlauts: äöüßÄÖÜ
    Chinese: 老师是四十四，是不是？
    Finnish: Mustan kissan paksut posket. (ah, no special chars) Kärpänen sanoi kärpäselle: tuu kattoon kattoon ku kaveri tapettiin tapettiin.
    """,
    language="en"
)
FIXT_ITEM = dict(
    title="Hello",
    link="http://www.holovaty.com/täst/",
    description="Testing.",
    content="Full content of our testing entry.",
    pubdate=NAIVE_DATE,
)

MINIMAL_FEED = {key: FIXT_FEED[key] for key in ("title", "link", "description")}
MINIMAL_ITEM = {key: FIXT_ITEM[key] for key in ("title", "link", "description")}


EXPECTED_RESULT_RSS = """<?xml version="1.0" encoding="utf-8"?>
<rss version="2.0" xmlns:atom="http://www.w3.org/2005/Atom"><channel><title>Poynter E-Media Tidbits</title><link>http://www.poynter.org/column.asp?id=31</link><description>A group Weblog by the sharpest minds in online media/journalism/publishing.
    Umlauts: äöüßÄÖÜ
    Chinese: 老师是四十四，是不是？
    Finnish: Mustan kissan paksut posket. (ah, no special chars) Kärpänen sanoi kärpäselle: tuu kattoon kattoon ku kaveri tapettiin tapettiin.
    </description><language>en</language><lastBuildDate>%DATE%</lastBuildDate><item><title>Hello</title><link>http://www.holovaty.com/t%C3%A4st/</link><description>Testing.</description><pubDate>Thu, 11 Aug 2016 00:00:00 -0000</pubDate></item></channel></rss>"""

EXPECTED_RESULT_ATOM = """<?xml version="1.0" encoding="utf-8"?>
<feed xml:lang="en" xmlns="http://www.w3.org/2005/Atom"><title>Poynter E-Media Tidbits</title><link href="http://www.poynter.org/column.asp?id=31" rel="alternate"/><id>http://www.poynter.org/column.asp?id=31</id><updated>%DATE%</updated><subtitle>A group Weblog by the sharpest minds in online media/journalism/publishing.
    Umlauts: äöüßÄÖÜ
    Chinese: 老师是四十四，是不是？
    Finnish: Mustan kissan paksut posket. (ah, no special chars) Kärpänen sanoi kärpäselle: tuu kattoon kattoon ku kaveri tapettiin tapettiin.
    </subtitle><entry><title>Hello</title><link href="http://www.holovaty.com/t%C3%A4st/" rel="alternate"/><published>2016-08-11T00:00:00Z</published><updated>2016-08-11T00:00:00Z</updated><id>tag:www.holovaty.com,2016-08-11:/t%C3%A4st/</id><summary type="html">Testing.</summary><content type="html">Full content of our testing entry.</content></entry></feed>"""

ENCODING = 'utf-8'

def build_expected_rss_result(feed, expected_result, encoding):
    # Result's date is of course different from the date in the fixture.
    # So make them equal!
    d = feedgenerator.rfc2822_date(feed.latest_post_date())
    s = expected_result.replace('%DATE%', d)
    if encoding:
        return s.encode(encoding)
    else:
        return s


def build_expected_atom_result(feed, expected_result, encoding):
    # Result's date is of course different from the date in the fixture.
    # So make them equal!
    d = feedgenerator.rfc3339_date(feed.latest_post_date())
    s = expected_result.replace('%DATE%', d)
    if encoding:
        return s.encode(encoding)
    else:
        return s


def test_000_types():
    for k, v in FIXT_FEED.items():
        assert isinstance(v, str)
    for k, v in FIXT_ITEM.items():
        if k == "pubdate" or k == "updateddate":
            assert isinstance(v, datetime.datetime)
        else:
            assert isinstance(v, str)
    assert isinstance(EXPECTED_RESULT_RSS, str)


def test_001_string_results_rss():
    #import ipdb; ipdb.set_trace()
    feed = feedgenerator.Rss201rev2Feed(**FIXT_FEED)
    feed.add_item(**FIXT_ITEM)
    result = feed.writeString(ENCODING)
    # On Python 3, result of feedgenerator is a unicode string!
    # So do not encode our expected_result.
    expected_result = build_expected_rss_result(feed, EXPECTED_RESULT_RSS, None)
    assert isinstance(result, type(expected_result))
    assert result == expected_result


def test_002_string_results_atom():
    #import ipdb; ipdb.set_trace()
    feed = feedgenerator.Atom1Feed(**FIXT_FEED)
    feed.add_item(**FIXT_ITEM)
    result = feed.writeString(ENCODING)
    # On Python 3, result of feedgenerator is a unicode string!
    # So do not encode our expected_result.
    expected_result = build_expected_atom_result(feed, EXPECTED_RESULT_ATOM, None)
    assert isinstance(result, type(expected_result))
    assert result == expected_result


@pytest.mark.parametrize("description, subtitle, fragment, nonfragment", [
    # Neither description nor subtitle are provided
    (None, None, None, "<subtitle/>"),
    ("", "", None, "<subtitle/>"),
    # Description is provided
    ("description", None, "<subtitle>description</subtitle>", None),
    ("description", "", "<subtitle>description</subtitle>", None),
    # Subtitle is provided
    (None, "subtitle", "<subtitle>subtitle</subtitle>", None),
    ("", "subtitle", "<subtitle>subtitle</subtitle>", None),
    # Both description & subtitle are provided; subtitle takes precedence
    ("description", "subtitle", "<subtitle>subtitle</subtitle>", "<subtitle>description</subtitle>"),
])
def test_subtitle(description, subtitle, fragment, nonfragment):
    """Test regression for https://github.com/getpelican/feedgenerator/issues/30.

    We test against all four possible combinations of description x
    subtitle parameters and additionally for None and "".

    description, subtitle are the values for the respective
    feed-parameters.

    fragment and nonfragment are text fragments that should be in the
    expected result or not.

    """
    FIXT_FEED = dict(
        title="title",
        link="https://example.com",
        description=description,
        subtitle=subtitle,
    )
    feed = feedgenerator.Atom1Feed(**FIXT_FEED)
    result = feed.writeString(ENCODING)
    if fragment:
        assert fragment in result
    if nonfragment:
        assert nonfragment not in result

@pytest.mark.parametrize("generator, date, expected_date_string", [
    (feedgenerator.Atom1Feed, AWARE_DATE, "2016-08-11T00:00:00-04:00"),
    (feedgenerator.Atom1Feed, AWARE_DATE_UTC, "2016-08-11T00:00:00+00:00"),
    (feedgenerator.Atom1Feed, NAIVE_DATE, "2016-08-11T00:00:00Z"),
    (feedgenerator.Rss201rev2Feed, AWARE_DATE, "Thu, 11 Aug 2016 00:00:00 -0400"),
    (feedgenerator.Rss201rev2Feed, AWARE_DATE_UTC, "Thu, 11 Aug 2016 00:00:00 +0000"),
    (feedgenerator.Rss201rev2Feed, NAIVE_DATE, "Thu, 11 Aug 2016 00:00:00 -0000"),
    (feedgenerator.RssUserland091Feed, AWARE_DATE, "Thu, 11 Aug 2016 00:00:00 -0400"),
    (feedgenerator.RssUserland091Feed, AWARE_DATE_UTC, "Thu, 11 Aug 2016 00:00:00 +0000"),
    (feedgenerator.RssUserland091Feed, NAIVE_DATE, "Thu, 11 Aug 2016 00:00:00 -0000"),
    ])
def test_timezone_handling(generator, date, expected_date_string):
    """
    Test that dates are handled correctly in all Feed generators.
    Also test special cases of no timezone given, vs timezone without offset
    """

    feed = generator(**FIXT_FEED)
    feed.add_item(**(FIXT_ITEM | {'pubdate': date}))
    result = feed.writeString(ENCODING)

    assert expected_date_string in result


def test_feed_default_properties():
    feed = feedgenerator.SyndicationFeed(**MINIMAL_FEED)

    for key in ("title", "link", "description"):
        assert feed.feed[key] == MINIMAL_FEED[key]

    for key in (
            "language",
            "author_email",
            "author_name",
            "author_link",
            "subtitle",
            "categories",
            "feed_url",
            "feed_copyright",
            "id",
            "ttl",
            "stylesheets",
    ):
        assert key in feed.feed

    # Testing this after testing that the property exists
    assert feed.feed["id"] == MINIMAL_FEED["link"]


def test_item_default_properties():
    feed = feedgenerator.SyndicationFeed(**MINIMAL_FEED)
    feed.add_item(**MINIMAL_ITEM)

    for key in ("title", "description"):
        expected_value = MINIMAL_ITEM[key]
        # The link in the test contains a multibyte character
        if key == "link":
            expected_value = quote(MINIMAL_ITEM[key], safe="/#%[]=:;$&()+,!?*@'~")
        assert feed.items[0][key] == expected_value

    for key in (
            "content",
            "author_email",
            "author_name",
            "author_link",
            "pubdate",
            "updateddate",
            "comments",
            "unique_id",
            "unique_id_is_permalink",
            "enclosures",
            "categories",
            "item_copyright",
            "ttl",
    ):
        assert key in feed.items[0]


def test_feed_kwargs():
    feed = feedgenerator.SyndicationFeed(test="test", **FIXT_FEED)
    assert "test" in feed.feed


def test_item_kwargs():
    feed = feedgenerator.SyndicationFeed(**FIXT_ITEM)
    feed.add_item(test="test", **FIXT_ITEM)

    assert "test" in feed.items[0]



from io import StringIO

ENCODING = 'utf-8'

S0 = 'hello world, Umlauts: äöüßÄÖÜ, Chinese: 四是四，十是十，十四是十四，四十是四十，四十四隻不識字之石獅子是死的'
S0_BYTES = 'fe fi foe fam'.encode(ENCODING)

#print("###", StringIO, "###")


def test_001_text():
    # If we throw unicode into the StringIO buffer, we'll
    # get unicode out of it.
    assert isinstance(S0, str)
    buf = StringIO()
    print(S0, file=buf, end="")
    s1 = buf.getvalue()
    assert isinstance(S0, type(s1))
    assert S0 == s1
    assert isinstance(s1, str)


def test_002_bytes():
    buf = StringIO()
    print(S0_BYTES, file=buf, end="")
    s1 = buf.getvalue()

    # In Python 3 StringIO *ALWAYS* returns str (=text=unicode) !
    # Even if we originally write bytes into the buffer, the value
    # we get out of it has type str!

    # Input is bytes
    assert isinstance(S0_BYTES, bytes)
    # Output is NOT bytes...
    assert not isinstance(S0_BYTES, type(s1))
    assert not isinstance(s1, bytes)
    # ...but str!
    assert isinstance(s1, str)
    # So the contents are not equal!
    assert S0_BYTES != s1
    # StringIO coerced bytes into str:
    # b'xyz' ---> "b'xyz'"
    assert str(S0_BYTES) == s1
    # See, the type info is literally present in the output str!
    assert "b'" + str(S0_BYTES, encoding=ENCODING) + "'" == s1
    # Coercion is NOT decoding!
    assert S0_BYTES.decode(ENCODING) != s1
    assert str(S0_BYTES, encoding=ENCODING) != s1
    # These are the same
    assert S0_BYTES.decode(ENCODING) == str(S0_BYTES, encoding=ENCODING)
    # Additional note:
    # If we do not specify an encoding when we create a StringIO
    # buffer, Python 3 automatically uses the locale's preferred
    # encoding: locale.getpreferredencoding()
    # Cf. http://docs.python.org/release/3.0.1/library/io.html#io.TextIOWrapper
    # In my case this is the same encoding as the encoding of this source file,
    # namely UTF-8. If on your system both encodings are different, you may
    # encounter other results than the above.
    #
    # In Python 3.2 the signature of StringIO() has changed. It is no more
    # possible to specify an encoding here.
